from typing import Dict

BALANCE_ENDPOINT = '/balance'


class Balance:
    """A balance model class"""

    def __init__(self, amount: float, currency: str) -> None:
        """Initialize a new Balance object

        :param amount: Balance amount
        :type amount: float
        :param currency: currency code
        :type currency: str
        """
        self._amount = amount
        self._currency = currency

    @property
    def amount(self) -> float:
        """Get the balance amount

        :return: amount
        :rtype: float
        """
        return self._amount

    @property
    def currency(self) -> str:
        """Get the currency code

        :return: currency code
        :rtype: str
        """
        return self._currency


class BalanceService:
    """A service for interacting with the balance endpoint"""

    def __init__(self, http_client) -> None:
        self._http_client = http_client

    def get(self) -> Balance:
        """Get the client's current balance

        :return: Balance object containing the amount and currency.
        :rtype: Balance
        """
        balance = self._http_client.get(BALANCE_ENDPOINT).json()
        return Balance(balance["amount"], balance["currency"])



from typing import List

STARTUP_SCRIPTS_ENDPOINT = '/scripts'


class StartupScript:
    """A startup script model class"""

    def __init__(self, id: str, name: str, script: str) -> None:
        """Initialize a new startup script object

        :param id: startup script id
        :type id: str
        :param name: startup script name
        :type name: str
        :param script: the actual script
        :type script: str
        """
        self._id = id
        self._name = name
        self._script = script

    @property
    def id(self) -> str:
        """Get the startup script id

        :return: startup script id
        :rtype: str
        """
        return self._id

    @property
    def name(self) -> str:
        """Get the startup script name

        :return: startup script name
        :rtype: str
        """
        return self._name

    @property
    def script(self) -> str:
        """Get the actual startup script code

        :return: startup script text
        :rtype: str
        """
        return self._script


class StartupScriptsService:
    """A service for interacting with the startup scripts endpoint"""

    def __init__(self, http_client) -> None:
        self._http_client = http_client

    def get(self) -> List[StartupScript]:
        """Get all of the client's startup scripts

        :return: list of startup script objects
        :rtype: List[StartupScript]
        """
        scripts = self._http_client.get(STARTUP_SCRIPTS_ENDPOINT).json()
        scripts_objects = list(map(lambda script: StartupScript(
            script['id'], script['name'], script['script']), scripts))
        return scripts_objects

    def get_by_id(self, id) -> StartupScript:
        """Get a specific startup script by id.

        :param id: startup script id
        :type id: str
        :return: startup script object
        :rtype: StartupScript
        """
        script = self._http_client.get(
            STARTUP_SCRIPTS_ENDPOINT + f'/{id}').json()[0]

        return StartupScript(script['id'], script['name'], script['script'])

    def delete(self, id_list: List[str]) -> None:
        """Delete multiple startup scripts by id

        :param id_list: list of startup scripts ids
        :type id_list: List[str]
        """
        payload = {"scripts": id_list}
        self._http_client.delete(STARTUP_SCRIPTS_ENDPOINT, json=payload)
        return

    def delete_by_id(self, id: str) -> None:
        """Delete a single startup script by id

        :param id: startup script id
        :type id: str
        """
        self._http_client.delete(STARTUP_SCRIPTS_ENDPOINT + f'/{id}')
        return

    def create(self, name: str, script: str) -> StartupScript:
        """Create a new startup script

        :param name: startup script name
        :type name: str
        :param script: startup script value
        :type script: str
        :return: the new startup script's id
        :rtype: str
        """
        payload = {"name": name, "script": script}
        id = self._http_client.post(STARTUP_SCRIPTS_ENDPOINT, json=payload).text
        return StartupScript(id, name, script)



import pytest
import responses  # https://github.com/getsentry/responses

from datacrunch.exceptions import APIException
from datacrunch.instances.instances import InstancesService, Instance
from datacrunch.constants import Actions, ErrorCodes

INVALID_REQUEST = ErrorCodes.INVALID_REQUEST
INVALID_REQUEST_MESSAGE = 'Your existence is invalid'

INSTANCE_ID = 'deadc0de-a5d2-4972-ae4e-d429115d055b'
SSH_KEY_ID = '12345dc1-a5d2-4972-ae4e-d429115d055b'

INSTANCE_TYPE = "1V100.6V"
INSTANCE_IMAGE = "fastai"
INSTANCE_HOSTNAME = "I'll be your host for today"
INSTANCE_DESCRIPTION = "hope you enjoy your GPU"
INSTANCE_STATUS = 'running'
INSTANCE_PRICE_PER_HOUR = 0.60
INSTANCE_LOCATION = 'FIN1'
INSTANCE_IP = '1.2.3.4'
INSTANCE_CREATED_AT = "whatchalookingatboy?"

PAYLOAD = [
    {
        "created_at": INSTANCE_CREATED_AT,
        "status": INSTANCE_STATUS,
        "ip": INSTANCE_IP,
        "cpu": {
            "description": "super-duper-cpu",
            "number_of_cores": 6
        },
        "gpu": {
            "description": "super-duper-gpu",
            "number_of_gpus": 1
        },
        "memory": {
            "description": "super-duper-memory",
            "size_in_gigabytes": 32
        },
        "storage": {
            "description": "super-duper-storage",
            "size_in_gigabytes": 320
        },
        "hostname": INSTANCE_HOSTNAME,
        "description": INSTANCE_DESCRIPTION,
        "location": INSTANCE_LOCATION,
        "price_per_hour": INSTANCE_PRICE_PER_HOUR,
        "instance_type": INSTANCE_TYPE,
        "image": INSTANCE_IMAGE,
        "id": INSTANCE_ID,
        "ssh_key_ids": [SSH_KEY_ID]
    }
]


class TestInstancesService:
    @pytest.fixture
    def instances_service(self, http_client):
        return InstancesService(http_client)

    @pytest.fixture
    def endpoint(self, http_client):
        return http_client._base_url + "/instances"

    def test_get_instances(self, instances_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.GET,
            endpoint,
            json=PAYLOAD,
            status=200
        )

        # act
        instances = instances_service.get()

        # assert
        assert type(instances) == list
        assert len(instances) == 1
        assert type(instances[0]) == Instance
        assert type(instances[0].ssh_key_ids) == list
        assert instances[0].id == INSTANCE_ID
        assert instances[0].ssh_key_ids == [SSH_KEY_ID]
        assert instances[0].status == INSTANCE_STATUS
        assert instances[0].image == INSTANCE_IMAGE
        assert instances[0].instance_type == INSTANCE_TYPE
        assert instances[0].price_per_hour == INSTANCE_PRICE_PER_HOUR
        assert instances[0].location == INSTANCE_LOCATION
        assert instances[0].description == INSTANCE_DESCRIPTION
        assert instances[0].hostname == INSTANCE_HOSTNAME
        assert instances[0].ip == INSTANCE_IP
        assert instances[0].created_at == INSTANCE_CREATED_AT
        assert type(instances[0].cpu) == dict
        assert type(instances[0].gpu) == dict
        assert type(instances[0].memory) == dict
        assert type(instances[0].storage) == dict
        assert responses.assert_call_count(endpoint, 1) is True

    def test_get_instances_by_status_successful(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint + "?status=running"
        responses.add(
            responses.GET,
            url,
            json=PAYLOAD,
            status=200
        )

        # act
        instances = instances_service.get(status='running')

        # assert
        assert type(instances) == list
        assert len(instances) == 1
        assert type(instances[0]) == Instance
        assert type(instances[0].ssh_key_ids) == list
        assert instances[0].id == INSTANCE_ID
        assert instances[0].ssh_key_ids == [SSH_KEY_ID]
        assert instances[0].status == INSTANCE_STATUS
        assert instances[0].image == INSTANCE_IMAGE
        assert instances[0].instance_type == INSTANCE_TYPE
        assert instances[0].price_per_hour == INSTANCE_PRICE_PER_HOUR
        assert instances[0].location == INSTANCE_LOCATION
        assert instances[0].description == INSTANCE_DESCRIPTION
        assert instances[0].hostname == INSTANCE_HOSTNAME
        assert instances[0].ip == INSTANCE_IP
        assert instances[0].created_at == INSTANCE_CREATED_AT
        assert type(instances[0].cpu) == dict
        assert type(instances[0].gpu) == dict
        assert type(instances[0].memory) == dict
        assert type(instances[0].storage) == dict
        assert responses.assert_call_count(url, 1) is True

    def test_get_instances_by_status_failed(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint + "?status=blabbering"
        responses.add(
            responses.GET,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            instances_service.get(status='blabbering')

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True

    def test_get_instance_by_id_successful(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/' + INSTANCE_ID
        responses.add(
            responses.GET,
            url,
            json=PAYLOAD[0],
            status=200
        )

        # act
        instance = instances_service.get_by_id(INSTANCE_ID)

        # assert
        assert type(instance) == Instance
        assert instance.id == INSTANCE_ID
        assert instance.ssh_key_ids == [SSH_KEY_ID]
        assert instance.status == INSTANCE_STATUS
        assert instance.image == INSTANCE_IMAGE
        assert instance.instance_type == INSTANCE_TYPE
        assert instance.price_per_hour == INSTANCE_PRICE_PER_HOUR
        assert instance.location == INSTANCE_LOCATION
        assert instance.description == INSTANCE_DESCRIPTION
        assert instance.hostname == INSTANCE_HOSTNAME
        assert instance.ip == INSTANCE_IP
        assert instance.created_at == INSTANCE_CREATED_AT
        assert type(instance.cpu) == dict
        assert type(instance.gpu) == dict
        assert type(instance.memory) == dict
        assert type(instance.storage) == dict
        assert responses.assert_call_count(url, 1) is True

    def test_get_instance_by_id_failed(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/x'
        responses.add(
            responses.GET,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            instances_service.get_by_id('x')

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True

    def test_create_instance_successful(self, instances_service, endpoint):
        # arrange - add response mock
        # create instance
        responses.add(
            responses.POST,
            endpoint,
            body=INSTANCE_ID,
            status=200
        )
        # get instance by id
        url = endpoint + '/' + INSTANCE_ID
        responses.add(
            responses.GET,
            url,
            json=PAYLOAD[0],
            status=200
        )

        # act
        instance = instances_service.create(
            instance_type=INSTANCE_TYPE,
            image=INSTANCE_IMAGE,
            ssh_key_ids=[SSH_KEY_ID],
            hostname=INSTANCE_HOSTNAME,
            description=INSTANCE_DESCRIPTION,
        )

        # assert
        assert type(instance) == Instance
        assert instance.id == INSTANCE_ID
        assert instance.ssh_key_ids == [SSH_KEY_ID]
        assert instance.status == INSTANCE_STATUS
        assert instance.image == INSTANCE_IMAGE
        assert instance.instance_type == INSTANCE_TYPE
        assert instance.price_per_hour == INSTANCE_PRICE_PER_HOUR
        assert instance.location == INSTANCE_LOCATION
        assert instance.description == INSTANCE_DESCRIPTION
        assert instance.hostname == INSTANCE_HOSTNAME
        assert instance.ip == INSTANCE_IP
        assert instance.created_at == INSTANCE_CREATED_AT
        assert type(instance.cpu) == dict
        assert type(instance.gpu) == dict
        assert type(instance.memory) == dict
        assert type(instance.storage) == dict
        assert responses.assert_call_count(endpoint, 1) is True
        assert responses.assert_call_count(url, 1) is True

    def test_create_instance_failed(self, instances_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.POST,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            instances_service.create(
                instance_type=INSTANCE_TYPE,
                image=INSTANCE_IMAGE,
                ssh_key_ids=[SSH_KEY_ID],
                hostname=INSTANCE_HOSTNAME,
                description=INSTANCE_DESCRIPTION,
            )

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_action_successful(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/action'
        responses.add(
            responses.POST,
            url,
            status=202
        )

        # act
        result = instances_service.action(id_list=[INSTANCE_ID], action=Actions.SHUTDOWN)

        # assert
        assert result is None
        assert responses.assert_call_count(url, 1) is True

    def test_action_failed(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/action'
        responses.add(
            responses.POST,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            instances_service.action(id_list=[INSTANCE_ID], action="fluxturcate")

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True

    def test_is_available_successful(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/availability/' + INSTANCE_TYPE
        responses.add(
            responses.GET,
            url,
            json=True,
            status=200
        )

        # act
        is_available = instances_service.is_available(INSTANCE_TYPE)

        # assert
        assert is_available is True
        assert responses.assert_call_count(url, 1) is True

    def test_is_available_failed(self, instances_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/availability/x'
        responses.add(
            responses.GET,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            instances_service.is_available('x')

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True



from typing import List

IMAGES_ENDPOINT = '/images'


class Image:
    """An image model class"""

    def __init__(self, id: str, name: str, image_type: str, details: List[str]) -> None:
        """Initialize an image object

        :param id: image id
        :type id: str
        :param name: image name
        :type name: str
        :param image_type: image type, e.g. 'ubuntu-20.04-cuda-11.0'
        :type image_type: str
        :param details: image details
        :type details: List[str]
        """
        self._id = id
        self._name = name
        self._image_type = image_type
        self._details = details

    @property
    def id(self) -> str:
        """Get the image id

        :return: image id
        :rtype: str
        """
        return self._id

    @property
    def name(self) -> str:
        """Get the image name

        :return: image name
        :rtype: str
        """
        return self._name

    @property
    def image_type(self) -> str:
        """Get the image type

        :return: image type
        :rtype: str
        """
        return self._image_type

    @property
    def details(self) -> List[str]:
        """Get the image details

        :return: image details
        :rtype: List[str]
        """
        return self._details


class ImagesService:
    """A service for interacting with the images endpoint"""

    def __init__(self, http_client) -> None:
        self._http_client = http_client

    def get(self) -> List[Image]:
        """Get the available instance images 

        :return: list of images objects
        :rtype: List[Image]
        """
        images = self._http_client.get(IMAGES_ENDPOINT).json()
        image_objects = list(map(lambda image: Image(
            image['id'], image['name'], image['image_type'], image['details']), images))
        return image_objects



import responses # https://github.com/getsentry/responses

from datacrunch.balance.balance import BalanceService, Balance


def test_balance(http_client):
    # arrange - add response mock
    responses.add(
        responses.GET,
        http_client._base_url + "/balance",
        json={"amount": 50.5, "currency": "usd"},
        status=200
    )

    balance_service = BalanceService(http_client)

    # act
    balance = balance_service.get()

    # assert
    assert type(balance) == Balance
    assert type(balance.amount) == float
    assert type(balance.currency) == str
    assert balance.amount == 50.5
    assert balance.currency == "usd"



from datacrunch.authentication.authentication import AuthenticationService
from datacrunch.balance.balance import BalanceService
from datacrunch.http_client.http_client import HTTPClient
from datacrunch.images.images import ImagesService
from datacrunch.instance_types.instance_types import InstanceTypesService
from datacrunch.instances.instances import InstancesService
from datacrunch.ssh_keys.ssh_keys import SSHKeysService
from datacrunch.startup_scripts.startup_scripts import StartupScriptsService
from datacrunch.constants import Actions, InstanceStatus, ErrorCodes
from datacrunch.__version__ import VERSION


class DataCrunchClient:
    """Client for interacting with DataCrunch's public API"""

    def __init__(self, client_id: str, client_secret: str, base_url: str = "https://api.datacrunch.io/v1") -> None:
        """The DataCrunch client

        :param client_id: client id
        :type client_id: str
        :param client_secret: client secret
        :type client_secret: str
        :param base_url: base url for all the endpoints, optional, defaults to "https://api.datacrunch.io/v1"
        :type base_url: str, optional
        """

        # Constants
        self.actions: Actions = Actions()
        """Available actions to perform on an instance"""

        self.instance_status: InstanceStatus = InstanceStatus()
        """Possible instance statuses"""

        self.error_codes: ErrorCodes = ErrorCodes()
        """Available error codes"""

        self.base_url: str = base_url
        """DataCrunch's Public API URL"""

        self.version: str = VERSION
        """Current SDK Version"""

        # Services
        self._authentication: AuthenticationService = AuthenticationService(
            client_id, client_secret, self.base_url)
        self._http_client: HTTPClient = HTTPClient(
            self._authentication, self.base_url)

        self.balance: BalanceService = BalanceService(self._http_client)
        """Balance service. Get client balance"""

        self.images: ImagesService = ImagesService(self._http_client)
        """Image service"""

        self.instance_types: InstanceTypesService = InstanceTypesService(
            self._http_client)
        """Instance type service"""

        self.instances: InstancesService = InstancesService(self._http_client)
        """Instances service. Deploy, delete, hibernate (etc) instances"""

        self.ssh_keys: SSHKeysService = SSHKeysService(self._http_client)
        """SSH keys service"""

        self.startup_scripts: StartupScriptsService = StartupScriptsService(
            self._http_client)
        """Startup Scripts service"""



class Actions:
    START = 'start'
    SHUTDOWN = 'shutdown'
    DELETE = 'delete'
    HIBERNATE = 'hibernate'
    RESTORE = 'restore'

    def __init__(self):
        return


class InstanceStatus:
    RUNNING = 'running'
    PROVISIONING = 'provisioning'
    OFFLINE = 'offline'
    STARTING_HIBERNATION = 'starting_hibernation'
    HIBERNATING = 'hibernating'
    RESTORING = 'restoring'
    ERROR = 'error'

    def __init__(self):
        return


class ErrorCodes:
    INVALID_REQUEST = "invalid_request"
    UNAUTHORIZED_REQUEST = "unauthorized_request"
    INSUFFICIENT_FUNDS = "insufficient_funds"
    FORBIDDEN_ACTION = "forbidden_action"
    NOT_FOUND = "not_found"
    SERVER_ERROR = "server_error"
    SERVICE_UNAVAILABLE = "service_unavailable"

    def __init__(self):
        return



import pytest
import responses  # https://github.com/getsentry/responses

from datacrunch.exceptions import APIException
from datacrunch.startup_scripts.startup_scripts import StartupScriptsService, StartupScript

INVALID_REQUEST = 'invalid_request'
INVALID_REQUEST_MESSAGE = 'Your existence is invalid'

SCRIPT_ID = 'deadc0de-a5d2-4972-ae4e-d429115d055b'
SCRIPT_NAME = 'next episode of _____'
SCRIPT_VALUE = 'this was not in the script!'

script_ID_2 = 'beefbeef-a5d2-4972-ae4e-d429115d055b'

PAYLOAD = [
    {
        'id': SCRIPT_ID,
        'name': SCRIPT_NAME,
        'script': SCRIPT_VALUE
    }
]


class TestStartupScripts:

    @pytest.fixture
    def startup_script_service(self, http_client):
        return StartupScriptsService(http_client)

    @pytest.fixture
    def endpoint(self, http_client):
        return http_client._base_url + "/scripts"

    def test_get_scripts(self, startup_script_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.GET,
            endpoint,
            json=PAYLOAD,
            status=200
        )

        # act
        scripts = startup_script_service.get()

        # assert
        assert type(scripts) == list
        assert len(scripts) == 1
        assert type(scripts[0]) == StartupScript
        assert scripts[0].id == SCRIPT_ID
        assert scripts[0].name == SCRIPT_NAME
        assert scripts[0].script == SCRIPT_VALUE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_get_script_by_id_successful(self, startup_script_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/' + SCRIPT_ID
        responses.add(
            responses.GET,
            url,
            json=PAYLOAD,
            status=200
        )

        # act
        script = startup_script_service.get_by_id(SCRIPT_ID)

        # assert
        assert type(script) == StartupScript
        assert script.id == SCRIPT_ID
        assert script.name == SCRIPT_NAME
        assert script.script == SCRIPT_VALUE
        assert responses.assert_call_count(url, 1) is True

    def test_get_script_by_id_failed(self, startup_script_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/x'
        responses.add(
            responses.GET,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            startup_script_service.get_by_id('x')

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True

    def test_create_script_successful(self, startup_script_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.POST,
            endpoint,
            body=SCRIPT_ID,
            status=201
        )

        # act
        script = startup_script_service.create(SCRIPT_NAME, SCRIPT_VALUE)

        # assert
        assert type(script) == StartupScript
        assert script.id == SCRIPT_ID
        assert responses.assert_call_count(endpoint, 1) is True

    def test_create_script_failed(self, startup_script_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.POST,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            startup_script_service.create(SCRIPT_NAME, SCRIPT_VALUE)

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_delete_scripts_successful(self, startup_script_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.DELETE,
            endpoint,
            status=200
        )

        # act
        result = startup_script_service.delete([SCRIPT_ID, script_ID_2])

        # assert
        assert result is None
        assert responses.assert_call_count(endpoint, 1) is True

    def test_delete_scripts_failed(self, startup_script_service, endpoint):
        # arrange - add response mock
        responses.add(
            responses.DELETE,
            endpoint,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            startup_script_service.delete(['x'])

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(endpoint, 1) is True

    def test_delete_script_by_id_successful(self, startup_script_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/' + SCRIPT_ID
        responses.add(
            responses.DELETE,
            url,
            status=200
        )

        # act
        result = startup_script_service.delete_by_id(SCRIPT_ID)

        # assert
        assert result is None
        assert responses.assert_call_count(url, 1) is True

    def test_delete_script_by_id_failed(self, startup_script_service, endpoint):
        # arrange - add response mock
        url = endpoint + '/x'
        responses.add(
            responses.DELETE,
            url,
            json={"code": INVALID_REQUEST, "message": INVALID_REQUEST_MESSAGE},
            status=400
        )

        # act
        with pytest.raises(APIException) as excinfo:
            startup_script_service.delete_by_id('x')

        # assert
        assert excinfo.value.code == INVALID_REQUEST
        assert excinfo.value.message == INVALID_REQUEST_MESSAGE
        assert responses.assert_call_count(url, 1) is True



import setuptools

version = {}
with open("datacrunch/__version__.py") as fp:
    exec(fp.read(), version)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="datacrunch",
    version=version['VERSION'],
    author="DataCrunch Oy",
    author_email="info@datacrunch.io",
    description="Official Python SDK for DataCrunch Public API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DataCrunch-io",
    packages=setuptools.find_packages(),
    install_requires=['requests>=2.25.1,<3'],
    extras_require={
        'dev': [''],
        'test': ['pytest>=6.2.1,<7', 'pytest-responses>=0.4.0,<1', 'responses>=0.12.1,<1']
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: English"
    ],
    python_requires='>=3.6',
)



import responses # https://github.com/getsentry/responses

from datacrunch.instance_types.instance_types import InstanceTypesService, InstanceType


def test_instance_types(http_client):
    # arrange - add response mock
    responses.add(
        responses.GET,
        http_client._base_url + "/instance-types",
        json=[
            {
                "id": "01cf5dc1-a5d2-4972-ae4e-d429115d055b",
                "cpu": {
                    "description": "48 CPU 3.5GHz",
                    "number_of_cores": 48
                },
                "gpu": {
                    "description": "8x NVidia Tesla V100",
                    "number_of_gpus": 8
                },
                "memory": {
                    "description": "192GB RAM",
                    "size_in_gigabytes": 192
                },
                "storage": {
                    "description": "1800GB NVME",
                    "size_in_gigabytes": 1800
                },
                "description": "Dedicated Bare metal Server",
                "pricePerHour": "5.00",
                "instance_type": "8V100.48M"
            }
        ],
        status=200
    )

    instance_types_service = InstanceTypesService(http_client)

    # act
    instance_types = instance_types_service.get()

    # assert
    assert type(instance_types) == list
    assert len(instance_types) == 1
    assert type(instance_types[0]) == InstanceType
    assert instance_types[0].id == '01cf5dc1-a5d2-4972-ae4e-d429115d055b'
    assert instance_types[0].description == "Dedicated Bare metal Server"
    assert instance_types[0].price_per_hour == 5.0
    assert instance_types[0].instance_type == "8V100.48M"
    assert type(instance_types[0].cpu) == dict
    assert type(instance_types[0].gpu) == dict
    assert type(instance_types[0].memory) == dict
    assert type(instance_types[0].storage) == dict
    assert instance_types[0].cpu['description'] == "48 CPU 3.5GHz"
    assert instance_types[0].gpu['description'] == "8x NVidia Tesla V100"
    assert instance_types[0].memory['description'] == "192GB RAM"
    assert instance_types[0].storage['description'] == "1800GB NVME"
    assert instance_types[0].cpu['number_of_cores'] == 48
    assert instance_types[0].gpu['number_of_gpus'] == 8
    assert instance_types[0].memory['size_in_gigabytes'] == 192
    assert instance_types[0].storage['size_in_gigabytes'] == 1800


from typing import List, Union

INSTANCES_ENDPOINT = '/instances'


class Instance:
    """An instance model class"""

    def __init__(self,
                 id: str,
                 instance_type: str,
                 image: str,
                 price_per_hour: float,
                 hostname: str,
                 description: str,
                 ip: str,
                 status: str,
                 created_at: str,
                 ssh_key_ids: List[str],
                 cpu: dict,
                 gpu: dict,
                 memory: dict,
                 storage: dict,
                 location: str = "FIN1",
                 startup_script_id: str = None
                 ) -> None:
        """Initialize the instance object

        :param id: instance id
        :type id: str
        :param instance_type: instance type. e.g. '8V100.48M'
        :type instance_type: str
        :param image: instance image type. e.g. 'ubuntu-20.04-cuda-11.0'
        :type image: str
        :param price_per_hour: price per hour
        :type price_per_hour: float
        :param hostname: instance hostname
        :type hostname: str
        :param description: instance description
        :type description: str
        :param ip: instance ip address
        :type ip: str
        :param status: instance current status, might be out of date if changed.
        :type status: str
        :param created_at: the time the instance was deployed (UTC)
        :type created_at: str
        :param ssh_key_ids: list of ssh keys ids
        :type ssh_key_ids: List[str]
        :param cpu: cpu details
        :type cpu: dict
        :param gpu: gpu details
        :type gpu: dict
        :param memory: memory details
        :type memory: dict
        :param storage: storate details
        :type storage: dict
        :param location: datacenter location, defaults to "FIN1"
        :type location: str, optional
        :param startup_script_id: startup script id, defaults to None
        :type startup_script_id: str, optional
        """
        self._id = id
        self._instance_type = instance_type
        self._image = image
        self._price_per_hour = price_per_hour
        self._location = location
        self._hostname = hostname
        self._description = description
        self._ip = ip
        self._status = status
        self._created_at = created_at
        self._ssh_key_ids = ssh_key_ids
        self._startup_script_id = startup_script_id
        self._cpu = cpu
        self._gpu = gpu
        self._memory = memory
        self._storage = storage

    @property
    def id(self) -> str:
        """Get the instance id

        :return: instance id
        :rtype: str
        """
        return self._id

    @property
    def instance_type(self) -> str:
        """Get the instance type

        :return: instance type
        :rtype: str
        """
        return self._instance_type

    @property
    def image(self) -> str:
        """Get the instance image type

        :return: instance image type
        :rtype: str
        """
        return self._image

    @property
    def price_per_hour(self) -> float:
        """Get the instance price per hour

        :return: price per hour
        :rtype: float
        """
        return self._price_per_hour

    @property
    def location(self) -> str:
        """Get the instance datacenter location

        :return: datacenter location
        :rtype: str
        """
        return self._location

    @property
    def hostname(self) -> str:
        """Get the instance hostname

        :return: hostname
        :rtype: str
        """
        return self._hostname

    @property
    def description(self) -> str:
        """Get the instance description

        :return: instance description
        :rtype: str
        """
        return self._description

    @property
    def ip(self) -> str:
        """Get the instance ip address

        :return: ip address
        :rtype: str
        """
        return self._ip

    @property
    def status(self) -> str:
        """Get the current instance status. might be out of date if changed.

        :return: instance status
        :rtype: str
        """
        return self._status

    @property
    def created_at(self) -> str:
        """Get the time when the instance was deployed (UTC)

        :return: time
        :rtype: str
        """
        return self._created_at

    @property
    def ssh_key_ids(self) -> List[str]:
        """Get the SSH key IDs of the instance

        :return: SSH key IDs
        :rtype: List[str]
        """
        return self._ssh_key_ids

    @property
    def startup_script_id(self) -> Union[str, None]:
        """Get the startup script ID or None if the is no script

        :return: startup script ID or None
        :rtype: Union[str, None]
        """
        return self._startup_script_id

    @property
    def cpu(self) -> dict:
        """Get the instance cpu details

        :return: cpu details
        :rtype: dict
        """
        return self._cpu

    @property
    def gpu(self) -> dict:
        """Get the instance gpu details

        :return: gpu details
        :rtype: dict
        """
        return self._gpu

    @property
    def memory(self) -> dict:
        """Get the instance memory details

        :return: memory details
        :rtype: dict
        """
        return self._memory

    @property
    def storage(self) -> dict:
        """Get the instance storage details

        :return: storage details
        :rtype: dict
        """
        return self._storage


class InstancesService:
    """A service for interacting with the instances endpoint"""

    def __init__(self, http_client) -> None:
        self._http_client = http_client

    def get(self, status: str = None) -> List[Instance]:
        """Get all of the client's non-deleted instances, or instances with specific status.

        :param status: optional, status of the instances, defaults to None
        :type status: str, optional
        :return: list of instance details objects
        :rtype: List[Instance]
        """
        instances_dict = self._http_client.get(
            INSTANCES_ENDPOINT, params={'status': status}).json()
        instances = list(map(lambda instance_dict: Instance(
            id=instance_dict['id'],
            instance_type=instance_dict['instance_type'],
            image=instance_dict['image'],
            price_per_hour=instance_dict['price_per_hour'],
            location=instance_dict['location'],
            hostname=instance_dict['hostname'],
            description=instance_dict['description'],
            ip=instance_dict['ip'],
            status=instance_dict['status'],
            created_at=instance_dict['created_at'],
            ssh_key_ids=instance_dict['ssh_key_ids'],
            startup_script_id=instance_dict['startup_script_id'] if 'startup_script_id' in instance_dict else None,
            cpu=instance_dict['cpu'],
            gpu=instance_dict['gpu'],
            memory=instance_dict['memory'],
            storage=instance_dict['storage']
        ), instances_dict))
        return instances

    def get_by_id(self, id: str) -> Instance:
        """Get an instance with specified id.

        :param id: instance id
        :type id: str
        :return: instance details object
        :rtype: Instance
        """
        instance_dict = self._http_client.get(
            INSTANCES_ENDPOINT + f'/{id}').json()
        instance = Instance(
            id=instance_dict['id'],
            instance_type=instance_dict['instance_type'],
            image=instance_dict['image'],
            price_per_hour=instance_dict['price_per_hour'],
            location=instance_dict['location'],
            hostname=instance_dict['hostname'],
            description=instance_dict['description'],
            ip=instance_dict['ip'],
            status=instance_dict['status'],
            created_at=instance_dict['created_at'],
            ssh_key_ids=instance_dict['ssh_key_ids'],
            startup_script_id=instance_dict['startup_script_id'] if 'startup_script_id' in instance_dict else None,
            cpu=instance_dict['cpu'],
            gpu=instance_dict['gpu'],
            memory=instance_dict['memory'],
            storage=instance_dict['storage']
        )
        return instance

    def create(self,
               instance_type: str,
               image: str,
               ssh_key_ids: list,
               hostname: str,
               description: str,
               location: str = "FIN1",
               startup_script_id: str = None) -> Instance:
        """Creates (deploys) a new instance

        :param instance_type: instance type. e.g. '8V100.48M'
        :type instance_type: str
        :param image: instance image type. e.g. 'ubuntu-20.04-cuda-11.0'
        :type image: str
        :param ssh_key_ids: list of ssh key ids
        :type ssh_key_ids: list
        :param hostname: instance hostname
        :type hostname: str
        :param description: instance description
        :type description: str
        :param location: datacenter location, defaults to "FIN1"
        :type location: str, optional
        :param startup_script_id: startup script id, defaults to None
        :type startup_script_id: str, optional
        :return: the new instance object
        :rtype: id
        """
        payload = {
            "instance_type": instance_type,
            "image": image,
            "ssh_key_ids": ssh_key_ids,
            "startup_script_id": startup_script_id,
            "hostname": hostname,
            "description": description,
            "location": location
        }
        id = self._http_client.post(INSTANCES_ENDPOINT, json=payload).text
        instance = self.get_by_id(id)
        return instance

    def action(self, id_list: Union[List[str], str], action: str) -> None:
        """Performs an action on a list of instances / single instance

        :param id_list: list of instance ids, or an instance id
        :type id_list: Union[List[str], str]
        :param action: the action to perform
        :type action: str
        """
        if type(id_list) is str:
            id_list = [id_list]

        payload = {
            "id": id_list,
            "action": action
        }

        self._http_client.post(INSTANCES_ENDPOINT + '/action', json=payload)
        return

    def is_available(self, instance_type: str) -> bool:
        """Returns True if a specific instance type is now available for deployment

        :param instance_type: instance type
        :type instance_type: str
        :return: True if available to deploy, False otherwise
        :rtype: bool
        """
        return self._http_client.get(INSTANCES_ENDPOINT + f'/availability/{instance_type}').json()
